@@ -1,9 +1,11 @@
 # stdlib
+from multimodal.text.wordembedding import get_dim_from_name
 import os
 import json
 from itertools import combinations
 from statistics import mean
 from collections import Counter
+from typing import List
 
 # librairies
 from tqdm import tqdm
@@ -19,9 +21,18 @@ from multimodal.datasets import vqa_utils
 from multimodal import DEFAULT_DATA_DIR
 from multimodal.utils import download_and_unzip
 from multimodal.datasets.vqa_utils import EvalAIAnswerProcessor
+from multimodal.text import WordEmbedding
 
 
-class VQA(Dataset):
+class AbstractVQA(Dataset):
+    def get_all_tokens(self) -> List:
+        raise NotImplementedError()
+
+    def evaluate(self, predictions) -> float:
+        raise NotImplementedError()
+
+
+class VQA(AbstractVQA):
 
     SPLITS = ["train", "val", "test", "test-dev"]
 
@@ -57,6 +68,7 @@ class VQA(Dataset):
         min_ans_occ=8,
         dir_features=None,
         label="multilabel",
+        tokenize_questions=False,
     ):
         """
         dir_data: dir for the multimodal cache (data will be downloaded in a vqa2/ folder inside this directory
@@ -66,6 +78,8 @@ class VQA(Dataset):
         label: either `multilabel`, or `best`. For `multilabel`, GT scores for questions are
             given by the score they are assigned by the VQA evaluation. 
             If `best`, GT is the label of the top answer.
+        tokenize_questions: If True, preprocessing will tokenize questions into tokens.
+            The tokens are stored in item["question_tokens"].
         """
         self.dir_data = dir_data
         if self.dir_data is None:
@@ -75,6 +89,9 @@ class VQA(Dataset):
         self.dir_features = dir_features or os.path.join(self.dir_data)
         self.label = label
         self.min_ans_occ = min_ans_occ
+        self.tokenize_questions = tokenize_questions
+        if self.tokenize_questions:
+            self.tokenizer = get_tokenizer("basic_english")
 
         # Test split has no annotations.
         self.has_annotations = self.split in self.url_annotations
@@ -84,7 +101,7 @@ class VQA(Dataset):
         self.dir_splits = {
             split: os.path.join(self.dir_dataset, split)
             for split in self.filename_questions.keys()
-        }
+        }  # vqa2/train/
 
         self.dir_splits["test-dev"] = "test"  # same directory
 
@@ -95,7 +112,7 @@ class VQA(Dataset):
         self.path_questions = {
             split: os.path.join(self.dir_splits[split], self.filename_questions[split])
             for split in self.filename_questions.keys()
-        }
+        }  # vqa2/train/OpenEnded_mscoco_train2014_questions.json
 
         # path download annotations
         self.path_original_annotations = {
@@ -103,7 +120,7 @@ class VQA(Dataset):
                 self.dir_splits[split], self.filename_annotations[split]
             )
             for split in self.filename_annotations.keys()
-        }
+        }  # vqa2/val/mscoco_val2014_annotations.json
 
         # processed annotations contain answer_token and answer scores
         self.processed_dirs = {
@@ -161,10 +178,33 @@ class VQA(Dataset):
 
     def get_all_tokens(self):
         tokenizer = get_tokenizer("basic_english")
-        return set(
-            (token for q in self.questions for token in tokenizer(q["question"]))
+        return list(
+            set((token for q in self.questions for token in tokenizer(q["question"])))
         )
 
+    def get_word_embeddings(self, name, freeze=True):
+        """
+        This will create the word embedding adapted to the dataset, and will cache it in the
+        dataset directory for faster usage.
+        """
+        os.makedirs(os.path.join(self.dir_dataset, "wordembeddings"), exist_ok=True)
+        path = os.path.join(self.dir_dataset, "wordembeddings", f"{name}.pth")
+        tokens = self.get_all_tokens()
+        if os.path.exists(path):
+            w_emb = WordEmbedding(tokens, dim=get_dim_from_name(name), freeze=freeze)
+            state_dict = torch.load(path)
+            w_emb.load_state_dict(state_dict)
+            print("done")
+        else:
+            print(
+                f"Loading and caching {name} word embeddings for {self.name} dataset."
+            )
+            w_emb = WordEmbedding.from_pretrained(
+                "glove.6B.300d", tokens=self.get_all_tokens(), dir_data=self.dir_data
+            )
+            torch.save(w_emb.state_dict(), path)
+        return w_emb
+
     def process_annotations(self):
         """Process answers to create answer tokens,
         and precompute VQA score for faster evaluation.
@@ -191,7 +231,7 @@ class VQA(Dataset):
                 # )
                 for ansDic in annot["answers"]:
                     ansDic["answer"] = processor(ansDic["answer"])
-
+            qid_to_scores = dict()
             print("\tPre-Computing answer scores")
             for annot in tqdm(all_annotations):
                 annot["scores"] = {}
@@ -204,12 +244,15 @@ class VQA(Dataset):
                         score = min(1, float(len(matching_ans)) / 3)
                         scores.append(score)
                     annot["scores"][ans] = mean(scores)
+                qid_to_scores[annot["question_id"]] = annot["scores"]
             print(f"Saving processed annotations at {path_train} and {path_val}")
 
             with open(self.path_annotations_processed["train"], "w") as f:
                 json.dump(annotations_train, f)
             with open(self.path_annotations_processed["val"], "w") as f:
                 json.dump(annotations_val, f)
+            with open(os.path.join(self.dir_dataset, "qid_to_scores.json"), "w") as f:
+                json.dump(qid_to_scores, f)
 
         #####################################
         # Processing min occurences of answer
@@ -281,33 +324,34 @@ class VQA(Dataset):
         }
         Aditionnaly, if visual features are used, keys from the features will be added.
         """
-        data = {"index": index}
-        data.update(self.questions[index])
+        item = {"index": index}
+        item.update(self.questions[index])
         if self.has_annotations:
-            data.update(self.annotations[index])
-
+            item.update(self.annotations[index])
             if self.label == "multilabel":
                 label = torch.zeros(len(self.answers))
                 for ans, score in self.annotations[index]["scores"].items():
                     if ans in self.ans_to_aid:
                         aid = self.ans_to_aid[ans]
                         label[aid] = score
-                data["label"] = label
+                item["label"] = label
             elif self.label == "best":
                 scores = self.annotations[index]["scores"]
                 best_ans = max(scores, key=scores.get)
                 ans_id = self.ans_to_aid[best_ans]
-                data["label"] = torch.tensor(ans_id)
+                item["label"] = torch.tensor(ans_id)
 
         if self.features is not None:
-            image_id = data["image_id"]
-            data.update(self.feats[image_id])
+            image_id = item["image_id"]
+            item.update(self.feats[image_id])
 
-        return data
+        if self.tokenize_questions:
+            item["question_tokens"] = self.tokenizer(item["question"])
+        return item
 
     @staticmethod
     def collate_fn(batch):
-        no_collate_keys = ["scores", "question_id"]
+        no_collate_keys = ["scores", "question_id", "question_tokens"]
         result_batch = {}
         for key in batch[0]:
             if key not in no_collate_keys:
@@ -329,29 +373,34 @@ class VQA(Dataset):
             ans_type = annot["answer_type"]
             scores["overall"].append(score)
             scores[ans_type].append(score)
-        return {key: mean(score_list) for key, score_list in scores.items()}
+        return {key: mean(score_list) if len(score_list) else 0.0 for key, score_list in scores.items()}
 
 
 class VQADataModule(pl.LightningDataModule):
+
+    dataset = VQA
+
     def __init__(
         self,
         dir_data: str,
         min_ans_occ=8,
         features=None,
+        tokenize_question=False,
         label="multilabel",
         batch_size=512,
         num_workers=4,
     ):
+        super().__init__()
         self.dir_data = dir_data
         self.label = label
         self.features = features
         self.min_ans_occ = min_ans_occ
         self.batch_size = batch_size
         self.num_workers = num_workers
-        super().__init__()
+        self.tokenize_question = tokenize_question
 
     def prepare_data(self):
-        VQA(
+        self.train_dataset = self.dataset(
             dir_data=self.dir_data,
             split="train",
             features=self.features,
@@ -359,46 +408,52 @@ class VQADataModule(pl.LightningDataModule):
             label=self.label,
         )
 
+    def setup(self):
+        self.val_dataset = self.dataset(
+            dir_data=self.dir_data,
+            split="val",
+            features=self.features,
+            min_ans_occ=self.min_ans_occ,
+            label=self.label,
+            tokenize_questions=self.tokenize_question,
+        )
+
     def train_dataloader(self):
         return DataLoader(
-            VQA(
-                dir_data=self.dir_data,
-                split="train",
-                features=self.features,
-                min_ans_occ=self.min_ans_occ,
-                label=self.label,
-            ),
+            self.train_dataset,
             batch_size=self.batch_size,
             shuffle=True,
-            collate_fn=VQA.collate_fn,
+            collate_fn=self.dataset.collate_fn,
             num_workers=self.num_workers,
+            pin_memory=True,
         )
 
     def val_dataloader(self):
         return DataLoader(
-            VQA(
-                dir_data=self.dir_data,
-                split="val",
-                min_ans_occ=self.min_ans_occ,
-                label=self.label,
-            ),
+            self.val_dataset,
             batch_size=self.batch_size,
             shuffle=False,
-            collate_fn=VQA.collate_fn,
+            collate_fn=self.dataset.collate_fn,
             num_workers=self.num_workers,
+            pin_memory=True,
         )
 
     def test_dataloader(self):
+        dataset = self.dataset(
+            dir_data=self.dir_data,
+            split="test",
+            features=self.features,
+            min_ans_occ=self.min_ans_occ,
+            label=self.label,
+            tokenize_questions=self.tokenize_question,
+        )
         return DataLoader(
-            VQA(
-                dir_data=self.dir_data,
-                split="test",
-                min_ans_occ=self.min_ans_occ,
-                label=self.label,
-            ),
+            dataset,
             batch_size=self.batch_size,
-            collate_fn=VQA.collate_fn,
+            shuffle=False,
+            collate_fn=self.dataset.collate_fn,
             num_workers=self.num_workers,
+            pin_memory=True,
         )
 
 
@@ -429,72 +484,8 @@ class VQA2(VQA):
     }
 
 
-class VQA2DataModule(pl.LightningDataModule):
-    def __init__(
-        self,
-        dir_data: str,
-        min_ans_occ=8,
-        features=None,
-        label="multilabel",
-        batch_size=512,
-    ):
-        self.dir_data = dir_data
-        self.label = label
-        self.features = features
-        self.min_ans_occ = min_ans_occ
-        self.batch_size = batch_size
-        super().__init__()
-
-    def prepare_data(self):
-        VQA2(
-            dir_data=self.dir_data,
-            split="train",
-            features=self.features,
-            min_ans_occ=self.min_ans_occ,
-            label=self.label,
-        )
-
-    def train_dataloader(self):
-        return DataLoader(
-            VQA2(
-                dir_data=self.dir_data,
-                split="train",
-                features=self.features,
-                min_ans_occ=self.min_ans_occ,
-                label=self.label,
-            ),
-            batch_size=self.batch_size,
-            shuffle=True,
-            collate_fn=VQA2.collate_fn,
-        )
-
-    def val_dataloader(self):
-        return DataLoader(
-            VQA2(
-                dir_data=self.dir_data,
-                split="val",
-                features=self.features,
-                min_ans_occ=self.min_ans_occ,
-                label=self.label,
-            ),
-            batch_size=self.batch_size,
-            shuffle=False,
-            collate_fn=VQA2.collate_fn,
-        )
-
-    def test_dataloader(self):
-        return DataLoader(
-            VQA2(
-                dir_data=self.dir_data,
-                split="test",
-                features=self.features,
-                min_ans_occ=self.min_ans_occ,
-                label=self.label,
-            ),
-            batch_size=self.batch_size,
-            shuffle=False,
-            collate_fn=VQA2.collate_fn,
-        )
+class VQA2DataModule(VQADataModule):
+    dataset = VQA2
 
 
 class VQACP(VQA):
@@ -533,69 +524,11 @@ class VQACP(VQA):
             self.annotations = json.load(f)
 
 
-class VQACPDataModule(pl.LightningDataModule):
-    def __init__(
-        self,
-        dir_data: str,
-        min_ans_occ=8,
-        features=None,
-        label="multilabel",
-        batch_size=512,
-    ):
-        self.dir_data = dir_data
-        self.label = label
-        self.features = features
-        self.min_ans_occ = min_ans_occ
-        self.batch_size = batch_size
-        super().__init__()
-
-    def prepare_data(self):
-        return VQACP(
-            dir_data=self.dir_data,
-            split="train",
-            features=self.features,
-            min_ans_occ=self.min_ans_occ,
-            label=self.label,
-        )
-
-    def train_dataloader(self):
-        return DataLoader(
-            VQACP(
-                dir_data=self.dir_data,
-                split="train",
-                features=self.features,
-                min_ans_occ=self.min_ans_occ,
-                label=self.label,
-            ),
-            batch_size=self.batch_size,
-            shuffle=True,
-            collate_fn=VQACP.collate_fn,
-        )
-
-    def val_dataloader(self):
-        return DataLoader(
-            VQACP(
-                dir_data=self.dir_data,
-                split="val",
-                min_ans_occ=self.min_ans_occ,
-                label=self.label,
-            ),
-            batch_size=self.batch_size,
-            shuffle=False,
-            collate_fn=VQACP.collate_fn,
-        )
+class VQACPDataModule(VQADataModule):
+    dataset = VQACP
 
     def test_dataloader(self):
-        return DataLoader(
-            VQACP(
-                dir_data=self.dir_data,
-                split="test",
-                min_ans_occ=self.min_ans_occ,
-                label=self.label,
-            ),
-            batch_size=self.batch_size,
-            collate_fn=VQACP.collate_fn,
-        )
+        return None
 
 
 class VQACP2(VQACP):
@@ -613,66 +546,8 @@ class VQACP2(VQACP):
     }
 
 
-class VQACP2DataModule(pl.LightningDataModule):
-    def __init__(
-        self,
-        dir_data: str,
-        min_ans_occ=8,
-        features=None,
-        label="multilabel",
-        batch_size=512,
-    ):
-        self.dir_data = dir_data
-        self.label = label
-        self.features = features
-        self.min_ans_occ = min_ans_occ
-        self.batch_size = batch_size
-        super().__init__()
-
-    def prepare_data(self):
-        VQACP2(
-            dir_data=self.dir_data,
-            split="train",
-            features=self.features,
-            min_ans_occ=self.min_ans_occ,
-            label=self.label,
-        )
-
-    def train_dataloader(self):
-        return DataLoader(
-            VQACP2(
-                dir_data=self.dir_data,
-                split="train",
-                features=self.features,
-                min_ans_occ=self.min_ans_occ,
-                label=self.label,
-            ),
-            batch_size=self.batch_size,
-            shuffle=True,
-            collate_fn=VQACP2.collate_fn,
-        )
-
-    def val_dataloader(self):
-        return DataLoader(
-            VQACP2(
-                dir_data=self.dir_data,
-                split="val",
-                min_ans_occ=self.min_ans_occ,
-                label=self.label,
-            ),
-            batch_size=self.batch_size,
-            shuffle=False,
-            collate_fn=VQACP2.collate_fn,
-        )
+class VQACP2DataModule(VQADataModule):
+    dataset = VQACP2
 
     def test_dataloader(self):
-        return DataLoader(
-            VQACP2(
-                dir_data=self.dir_data,
-                split="test",
-                min_ans_occ=self.min_ans_occ,
-                label=self.label,
-            ),
-            batch_size=self.batch_size,
-            collate_fn=VQACP2.collate_fn,
-        )
+        return None
